﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Collections;

namespace XrProjection
{
    public partial class Form1 : Form
    {
        private string[] CoordinateSystem = new string[] {
            "Bessel 타원체의 경위도",
            "WGS84 타원체의 경위도",
            "GRS80 타원체의 경위도",
            "카텍(Katec)",
            "대한민국 TM 서부원점(Bessel 타원체) - 10.405 보정",
            "대한민국 TM 중부원점(Bessel 타원체) - 10.405 보정",
            "대한민국 TM 동부원점(Bessel 타원체) - 10.405 보정",
            "대한민국 TM 동해원점(Bessel 타원체) - 10.405 보정",
            "대한민국 TM 재주도원점(Bessel 타원체)",
            "대한민국 TM 서부원점(GRS80 타원체)",
            "대한민국 TM 중부원점(GRS80 타원체)",
            "대한민국 TM 동부원점(GRS80 타원체)",
            "대한민국 TM 동해원점(GRS80 타원체)",
            "UTM 51 Zone(WGS80 타원체)",
            "UTM 52 Zone(WGS80 타원체)",
            "UTM-K(GRS80 타원체)",
            "UTM-K(Bessel 타원체)",
            "EPGS:900913 / EPSG:3875"
        };

        private string[] Proj4String = new string[] {
            "+proj=longlat +ellps=bessel +no_defs",
            "+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs",
            "+proj=longlat +ellps=GRS80 +no_defs",
            "+proj=tmerc +lat_0=38N +lon_0=128E +ellps=bessel +x_0=400000 +y_0=600000 +k=0.9999 +units=m +no_defs",
            "+proj=tmerc +lat_0=38N +lon_0=125.00289027E +ellps=bessel +x_0=200000 +y_0=500000 +k=1 +units=m +no_defs",
            "+proj=tmerc +lat_0=38N +lon_0=127.00289027E +ellps=bessel +x_0=200000 +y_0=500000 +k=1 +units=m +no_defs",
            "+proj=tmerc +lat_0=38N +lon_0=129.00289027E +ellps=bessel +x_0=200000 +y_0=500000 +k=1 +units=m +no_defs",
            "+proj=tmerc +lat_0=38N +lon_0=131.00289027E +ellps=bessel +x_0=200000 +y_0=500000 +k=1 +units=m +no_defs",
            "+proj=tmerc +lat_0=38N +lon_0=127.00289027E +ellps=bessel +x_0=200000 +y_0=550000 +k=1 +units=m +no_defs",
            "+proj=tmerc +lat_0=38N +lon_0=125E +ellps=GRS80 +x_0=200000 +y_0=600000 +k=1 +units=m +no_defs",
            "+proj=tmerc +lat_0=38N +lon_0=127E +ellps=GRS80 +x_0=200000 +y_0=600000 +k=1 +units=m +no_defs",
            "+proj=tmerc +lat_0=38N +lon_0=129E +ellps=GRS80 +x_0=200000 +y_0=600000 +k=1 +units=m +no_defs",
            "+proj=tmerc +lat_0=38N +lon_0=131E +ellps=GRS80 +x_0=200000 +y_0=600000 +k=1 +units=m +no_defs",
            "+proj=utm +zone=51 +ellps=WGS84 +datum=WGS84 +units=m +no_defs",
            "+proj=utm +zone=52 +ellps=WGS84 +datum=WGS84 +units=m +no_defs",
            "+proj=tmerc +lat_0=38N +lon_0=127.50289E +ellps=GRS80 +x_0=1000000 +y_0=2000000 +k=0.9996 +units=m +no_defs",
            "+proj=tmerc +lat_0=38N +lon_0=127.50289E +ellps=bessel +x_0=1000000 +y_0=2000000 +k=0.9996 +units=m +no_defs",
            //"+proj=merc +a=6378137 +b=6378137 +lat_ts=0.0 +lon_0=0.0 +x_0=0.0 +y_0=0 +k=1.0 +units=m +nadgrids=@null +no_defs"
            "+proj=merc +a=6378137 +b=6378137 +lat_ts=0.0 +lon_0=0.0 +x_0=0.0 +y_0=0 +k=1.0 +units=m +nadgrids=@null +no_defs"
        };

        bool bInputValueIsDegree = false;
        bool bOutputValueIsDegree = false;
        
        public Form1()
        {
            InitializeComponent();

            for (int iCoordSys = 0; iCoordSys < CoordinateSystem.Length; ++iCoordSys)
            {
                cbProjectionSource1.Items.Add(CoordinateSystem[iCoordSys]);
                cbProjectionTarget1.Items.Add(CoordinateSystem[iCoordSys]);
                cbProjectionSource2.Items.Add(CoordinateSystem[iCoordSys]);
                cbProjectionTarget2.Items.Add(CoordinateSystem[iCoordSys]);
                cbProjectionSource3.Items.Add(CoordinateSystem[iCoordSys]);
                cbProjectionTarget3.Items.Add(CoordinateSystem[iCoordSys]);
            }

            cbProjectionSource1.SelectedIndex = -1;
            cbProjectionTarget1.SelectedIndex = -1;
            cbProjectionSource2.SelectedIndex = -1;
            cbProjectionTarget2.SelectedIndex = -1;
            cbProjectionSource3.SelectedIndex = -1;
            cbProjectionTarget3.SelectedIndex = -1;

            rb3Params_CheckedChanged(null, null);
        }

        private void button1_Click(object sender, EventArgs e)
        {
            DialogResult DR = openFileDialog.ShowDialog();
            if(DR == DialogResult.OK)
            {
                tbInputSHP.Text = openFileDialog.FileName;
            }
        }

        private void button2_Click(object sender, EventArgs e)
        {
            DialogResult DR = saveFileDialog.ShowDialog();
            if(DR == DialogResult.OK)
            {
                tbOutputSHP.Text = saveFileDialog.FileName;
            }
        }

        private void btnClose_Click(object sender, EventArgs e)
        {
            Close();
        }

        private String getProj4String(int Index)
        {
            String toWGS84 = "";
            if (cbUseParams.Checked && (Index != 1)) // WGS84 타원체에는 붙이지 않음
            {
                if (rb3Params.Checked)
                {
                    toWGS84 = " +towgs84=" + tbDX.Text + "," + tbDY.Text + "," + tbDZ.Text;
                }
                else
                {
                    toWGS84 = " +towgs84=" + tbDX.Text + "," + tbDY.Text + "," + tbDZ.Text + "," + tbOMEGA.Text + "," + tbPHI.Text + "," + tbKAPPA.Text + "," + tbDS.Text;
                }
            }

            //MessageBox.Show(toWGS84);

            return Proj4String[Index] + toWGS84;
        }

        private bool SetupProjection(XrMapLib.Projection proj)
        {
            int IndexTarget = cbProjectionTarget1.SelectedIndex;
            if (IndexTarget == -1)
            {
                MessageBox.Show("출력 SHP의 좌표계를 지정해주세요.");
                return false;
            }

            int IndexSource = cbProjectionSource1.SelectedIndex;
            if (IndexSource == -1)
            {
                MessageBox.Show("입력 SHP의 좌표계를 지정해주세요.");
                return false;
            }

            proj.SourceProj4String = getProj4String(IndexSource);
            proj.TargetProj4String = getProj4String(IndexTarget);

            bInputValueIsDegree = proj.SourceProj4String.Contains("longlat");
            bOutputValueIsDegree = proj.TargetProj4String.Contains("longlat");

            return true;
        }

        private bool ProjectSHPFile(String InputSHPFileName, String OutputSHPFileName)
        {
            XrMapLib.Projection proj = new XrMapLib.Projection();

            SetupProjection(proj);

            XrMapLib.Coordinate coord = new XrMapLib.Coordinate();

            // Add Input Layer into OpenMap-Xr
            bool bExisted = axXr1.Layers.IsLayerExist("map");
            if (bExisted) axXr1.Layers.RemoveLayer("map");

            if (tbInputSHP.Text.Trim().Length == 0 || tbOutputSHP.Text.Trim().Length == 0 || !System.IO.File.Exists(InputSHPFileName))
            {
                MessageBox.Show("SHP 파일을 옳바르게 지정해주세요.");
                return false;
            }

            axXr1.Layers.AddShapeMapLayer("map", InputSHPFileName);
            axXr1.WaitForAllConnections();

            XrMapLib.IShapeMapLayer lyr = axXr1.Layers.GetLayerAsShapeMap("map");
            XrMapLib.XrShapeTypeEnum shapeType = lyr.ShapeType;
            XrMapLib.IShapeTable Tbl = lyr.ShapeTable;
            XrMapLib.Tool tool = new XrMapLib.Tool();
            if (!OutputSHPFileName.ToLower().EndsWith(".shp"))
            {
                OutputSHPFileName += ".shp";
            }

            XrMapLib.EsriShpFileHandler SFH = tool.CreateEsriShpFileHandler(OutputSHPFileName, Tbl.ShapeType, null);
            if (SFH == null)
            {
                MessageBox.Show("Error: Creating shp file");
                return false;
            }

            int CountRows = Tbl.RowCount;
            XrMapLib.IShapeRow row = null;
            XrMapLib.CoordinateListCollection clc = new XrMapLib.CoordinateListCollection();
            XrMapLib.CoordinateList cl = new XrMapLib.CoordinateList();
            progressBar.Maximum = CountRows;
            progressBar.Value = 0;

            for (int FID = 0; FID < CountRows; ++FID)
            {
                progressBar.Value++;
                row = Tbl.GetRow(FID);
                if (row.Load())
                {
                    int cntParts = row.PartCount;
                    bool bPartCountZero = (cntParts == 0);
                    if (bPartCountZero)
                    {
                        MessageBox.Show("Null Shape !");
                        return false;
                    }

                    if (shapeType == XrMapLib.XrShapeTypeEnum.XrPoint)
                    {
                        coord = row.GetVertex(0);

                        if (bInputValueIsDegree)
                        {
                            coord.X = proj.DEG2RAD(coord.X);
                            coord.Y = proj.DEG2RAD(coord.Y);
                        }

                        proj.Transform(coord);

                        if (bOutputValueIsDegree)
                        {
                            coord.X = proj.RAD2DEG(coord.X);
                            coord.Y = proj.RAD2DEG(coord.Y);
                        }

                        SFH.AddPoint(coord.X, coord.Y, null);
                    }
                    else
                    {
                        clc.Clear();
                        for (int iPart = 0; iPart < cntParts; ++iPart)
                        {
                            int cntVertex = row.GetVertexCountFromPart(iPart);
                            cl.Clear();
                            for (int iVertex = 0; iVertex < cntVertex; ++iVertex)
                            {
                                if (bPartCountZero)
                                    coord = row.GetVertex(iVertex);
                                else
                                    coord = row.GetVertexFromPart(iPart, iVertex);

                                if (bInputValueIsDegree)
                                {
                                    coord.X = proj.DEG2RAD(coord.X);
                                    coord.Y = proj.DEG2RAD(coord.Y);
                                }

                                proj.Transform(coord);

                                if (bOutputValueIsDegree)
                                {
                                    coord.X = proj.RAD2DEG(coord.X);
                                    coord.Y = proj.RAD2DEG(coord.Y);
                                }

                                cl.Add(coord.X, coord.Y);
                            }
                            clc.Add(cl);
                        }

                        SFH.AddMultiPart(clc, null);
                    }

                    row.Unload();
                }
                else
                {
                    SFH.Close();
                    MessageBox.Show("Input SHP is broken.");
                    return false;
                }
            }

            SFH.Close();

            string SourceFileDBF = InputSHPFileName.ToLower().Replace(".shp", ".dbf");
            string DestFileDBF = OutputSHPFileName.ToLower().Replace(".shp", ".dbf");
            System.IO.File.Copy(SourceFileDBF, DestFileDBF, true);

            return true;
        }

        private void btnDo_Click(object sender, EventArgs e)
        {
            string InputSHPFileName = tbInputSHP.Text.ToLower();
            string OutputSHPFileName = tbOutputSHP.Text;

            if (ProjectSHPFile(InputSHPFileName, OutputSHPFileName))
            {
                MessageBox.Show("Completed");
            }
        }

        private void btnDoSingle_Click(object sender, EventArgs e)
        {
            XrMapLib.Projection proj = new XrMapLib.Projection();

            if (!SetupProjection(proj)) return;

            XrMapLib.Coordinate coord = new XrMapLib.Coordinate();
            double InX, InY;
            if(!Double.TryParse(tbInputCoordX.Text, out InX) || !Double.TryParse(tbInputCoordY.Text, out InY))
            {
                MessageBox.Show("입력된 좌표가 옳바르지 않습니다.");
                return;
            }

            coord.X = InX;
            coord.Y = InY;

            if (bInputValueIsDegree)
            {
                coord.X = proj.DEG2RAD(coord.X);
                coord.Y = proj.DEG2RAD(coord.Y);
            }

            proj.Transform(coord);

            if (bOutputValueIsDegree)
            {
                coord.X = proj.RAD2DEG(coord.X);
                coord.Y = proj.RAD2DEG(coord.Y);
            }

            tbOutputCoordX.Text = coord.X.ToString();
            tbOutputCoordY.Text = coord.Y.ToString();
        }

        private void rb3Params_CheckedChanged(object sender, EventArgs e)
        {
            tbOMEGA.Enabled = !rb3Params.Checked;
            tbPHI.Enabled = !rb3Params.Checked;
            tbKAPPA.Enabled = !rb3Params.Checked;
            tbDS.Enabled = !rb3Params.Checked;
        }

        private void btnDoDMSToDegree_Click(object sender, EventArgs e)
        {
            try
            {
                double d = Double.Parse(tbD.Text);
                double m = Double.Parse(tbM.Text);
                double s = Double.Parse(tbS.Text);

                tbDegree.Text = (d + m / 60.0 + s / 3600.0).ToString();
            }
            catch (Exception)
            {
                MessageBox.Show("옳바른 값을 입력해주세요.");
            }
        }

        private void btnDoDegreeToDMS_Click(object sender, EventArgs e)
        {
            try
            {
                double degree = Double.Parse(tbDegree.Text);

                double d = (int)degree;
                double i_d = (int)d;

                double m = (degree - d) * 60;
                double i_m = (int)m;

                double s = (m - i_m) * 60;
                double i_s = (int)s;

                tbD.Text = ((int)i_d).ToString();
                tbM.Text = ((int)i_m).ToString();
                tbS.Text = s.ToString();
            }
            catch (Exception)
            {
                MessageBox.Show("옳바른 값을 입력해주세요.");
            }
        }

        private void cbUseParams_CheckedChanged(object sender, EventArgs e)
        {
            this.rb3Params.Enabled = cbUseParams.Checked;
            this.rb7Params.Enabled = cbUseParams.Checked;
            this.tbKAPPA.Enabled = cbUseParams.Checked;
            this.tbOMEGA.Enabled = cbUseParams.Checked;
            this.tbPHI.Enabled = cbUseParams.Checked;
            this.tbDS.Enabled = cbUseParams.Checked;
            this.tbDX.Enabled = cbUseParams.Checked;
            this.tbDY.Enabled = cbUseParams.Checked;
            this.tbDZ.Enabled = cbUseParams.Checked;
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            cbUseParams_CheckedChanged(null, null);
        }

        private void button4_Click(object sender, EventArgs e)
        {
            DialogResult DR = folderBrowserDialog.ShowDialog();
            if (DR == DialogResult.OK)
            {
                tbInputFolder.Text = folderBrowserDialog.SelectedPath;
            }
        }

        private void button5_Click(object sender, EventArgs e)
        {
            DialogResult DR = folderBrowserDialog.ShowDialog();
            if (DR == DialogResult.OK)
            {
                tbOutputFolder.Text = folderBrowserDialog.SelectedPath;
            }
        }

        private void GetShpFileNames(string folderName, ArrayList fileNamesList)
        {
            System.IO.DirectoryInfo di = new System.IO.DirectoryInfo(folderName);
            foreach (System.IO.FileInfo f in di.GetFiles())
            {
                if (f.Extension.ToLower().CompareTo(".shp") == 0)
                {
                    String strInFileName = di.FullName + "\\" + f.Name;
                    fileNamesList.Add(strInFileName);
                }
            }

            foreach (System.IO.DirectoryInfo sd in di.GetDirectories())
            {
                GetShpFileNames(sd.FullName, fileNamesList);
            }
        }

        private void btnDoFolder_Click(object sender, EventArgs e)
        {
            ArrayList fileNamesList = new ArrayList();

            String strInFolder = tbInputFolder.Text.Trim();
            String strOutFolder = tbOutputFolder.Text.Trim();

            if (strInFolder.Length == 0 || strOutFolder.Length == 0)
            {
                MessageBox.Show("작업할 폴더를 지정하세요!");
                return;
            }

            GetShpFileNames(strInFolder, fileNamesList);
            int cntFileNames = fileNamesList.Count;

            progressBarFolder.Maximum = cntFileNames;
            progressBarFolder.Value = 0;

            for (int i = 0; i < cntFileNames; i++)
            {
                progressBarFolder.Value++;
                String strInFileName = fileNamesList[i].ToString();
                String strOutFileName = strInFileName.Replace(strInFolder, strOutFolder);

                string FileNameOnly = System.IO.Path.GetFileName(strOutFileName);
                String strOutFolderName = strOutFileName.Replace("\\" + FileNameOnly, "");

                if (!System.IO.Directory.Exists(strOutFolderName))
                {
                    System.IO.Directory.CreateDirectory(strOutFolderName);
                }

                tbInputSHP.Text = strInFileName;
                tbOutputSHP.Text = strOutFileName;
                if (!ProjectSHPFile(strInFileName, strOutFileName))
                {
                    MessageBox.Show("변환 실패 : " + strInFileName);
                    return;
                }
            }

            MessageBox.Show("Completed");
        }

        private void cbProjectionSource1_SelectedIndexChanged(object sender, EventArgs e)
        {
            cbProjectionSource2.SelectedIndex = cbProjectionSource1.SelectedIndex;
            cbProjectionSource3.SelectedIndex = cbProjectionSource1.SelectedIndex;
        }

        private void cbProjectionSource2_SelectedIndexChanged(object sender, EventArgs e)
        {
            cbProjectionSource1.SelectedIndex = cbProjectionSource2.SelectedIndex;
            cbProjectionSource3.SelectedIndex = cbProjectionSource2.SelectedIndex;
        }

        private void cbProjectionSource3_SelectedIndexChanged(object sender, EventArgs e)
        {
            cbProjectionSource1.SelectedIndex = cbProjectionSource3.SelectedIndex;
            cbProjectionSource2.SelectedIndex = cbProjectionSource3.SelectedIndex;
        }

        private void cbProjectionTarget1_SelectedIndexChanged(object sender, EventArgs e)
        {
            cbProjectionTarget2.SelectedIndex = cbProjectionTarget1.SelectedIndex;
            cbProjectionTarget3.SelectedIndex = cbProjectionTarget1.SelectedIndex;
        }

        private void cbProjectionTarget2_SelectedIndexChanged(object sender, EventArgs e)
        {
            cbProjectionTarget1.SelectedIndex = cbProjectionTarget2.SelectedIndex;
            cbProjectionTarget3.SelectedIndex = cbProjectionTarget2.SelectedIndex;
        }

        private void cbProjectionTarget3_SelectedIndexChanged(object sender, EventArgs e)
        {
            cbProjectionTarget1.SelectedIndex = cbProjectionTarget3.SelectedIndex;
            cbProjectionTarget2.SelectedIndex = cbProjectionTarget3.SelectedIndex;
        }
    }
}